@@ -107,7 +107,7 @@ unqueue_command (MPProcess *process, MIInfo *info, StackElement **stack)
  * Block Manipulation
  ************************************************************/
 
-static void
+void
 block_unref (Block *block)
 {
 	block->refcount--;
@@ -144,11 +144,13 @@ prepare_map (Map *map)
 		result = g_array_new (FALSE, FALSE, sizeof(Symbol));
 
 		for (i = 0; i < map->symcount; i++) {
-			if (map->syms[i]->flags & BSF_FUNCTION)	{
+			/* FIXME: is strcmp() agaings ".text" really right? */
+			if ((map->syms[i]->flags & BSF_FUNCTION) &&
+			    (strcmp (map->syms[i]->section->name, ".text") == 0))
+			{
 				symbol.addr = bfd_asymbol_value(map->syms[i]);
 				symbol.size = 0;
 				symbol.name = demangle (map, bfd_asymbol_name(map->syms[i]));
-	      
 				g_array_append_vals (result, &symbol, 1);
 			}
 		}
@@ -158,6 +160,12 @@ prepare_map (Map *map)
 		qsort (result->data, result->len, sizeof(Symbol), compare_address);
   
 		map->symbols =result;
+
+		/* Check for position independent executable */
+		if (!map->do_offset && map->symbols->len &&
+		    map->addr > ((Symbol*)map->symbols->data)[map->symbols->len - 1].addr) {
+			map->do_offset = TRUE;
+		}
 	}
 }
 
@@ -182,7 +190,7 @@ process_read_maps (MPProcess *process)
 	}
 
 	while (fgets(buffer, 1023, in)) {
-		int count = sscanf (buffer, "%x-%x %15s %*x %u:%u %u %255s",
+		int count = sscanf (buffer, "%x-%x %15s %*x %x:%x %u %255s",
 				    &start, &end, perms, &major, &minor, &inode, file);
 		if (count >= 6)	{
 			if (strcmp (perms, "r-xp") == 0) {
@@ -278,7 +286,7 @@ locate_map (MPProcess *process, guint addr)
 	return map;
 }
 
-Symbol *
+const Symbol *
 process_locate_symbol (MPProcess *process, guint addr)
 {
 	Symbol *data;
@@ -291,7 +299,7 @@ process_locate_symbol (MPProcess *process, guint addr)
 		return NULL;
 
 	if (!map->symbols || (map->symbols->len == 0))
-		return NULL;
+		return (Symbol*)map;
   
 	if (map->do_offset)
 		addr -= map->addr;
@@ -396,6 +404,16 @@ process_sections (MPProcess *process, SectionFunc func, gpointer user_data)
  * Communication with subprocess
  ************************************************************/
 
+static GHashTable *
+get_block_table (MPProcess *process)
+{
+	if (!process->block_table)
+		process->block_table = g_hash_table_new (g_direct_hash, NULL);
+
+	return process->block_table;
+		
+}
+
 static void
 process_free_block (gpointer key, gpointer value, gpointer data)
 {
@@ -418,9 +436,13 @@ process_duplicate (MPProcess *process)
 {
 	MPProcess *new_process = process_new (process->server);
 
-	g_hash_table_foreach (process->block_table,
-			      process_duplicate_block,
-			      new_process->block_table);
+	if (process->block_table) {
+		GHashTable *new_block_table = get_block_table(new_process);
+
+		g_hash_table_foreach (process->block_table,
+				      process_duplicate_block,
+				      new_block_table);
+	}
 
 	new_process->bytes_used = process->bytes_used;
 	new_process->n_allocations = process->n_allocations;
@@ -472,16 +494,6 @@ process_reinit (MPProcess *process)
 	process->command_queue = NULL;
 }
 
-static GHashTable *
-get_block_table (MPProcess *process)
-{
-	if (!process->block_table)
-		process->block_table = g_hash_table_new (g_direct_hash, NULL);
-
-	return process->block_table;
-		
-}
-
 static StackStash *
 get_stack_stash (MPProcess *process)
 {
@@ -524,6 +536,12 @@ process_command (MPProcess *process, MIInfo *info, StackElement *stack)
 	case MI_EXIT:
 		/* Handled before, ignore */
 		break;
+
+	case MI_TIME:
+		info->alloc.old_ptr = NULL;
+		info->alloc.new_ptr = (void *)process->seqno;
+		info->alloc.size = 1;
+		/* Fall through */
 		
 	default: /* MALLOC / REALLOC / FREE */
 		block_table = get_block_table (process);
@@ -581,46 +599,57 @@ input_func (GIOChannel  *source,
 	MPProcess *input_process = data;
 	MPProcess *process = NULL;
   
-	g_io_channel_read (source, (char *)&info, sizeof(info), &count);
+	do {
+		memset (&info, 0, sizeof (info));
+		info.operation = -1;
+		count = 0;
+		if (g_io_channel_get_flags(source) & G_IO_FLAG_IS_READABLE)
+			g_io_channel_read_chars (source, (char *)&info, sizeof(info), &count, NULL);
 
-	if (count == 0) {
-		g_io_channel_unref (input_process->input_channel);
-		input_process->input_channel = NULL;
+		if (count == 0) {
+			g_io_channel_unref (input_process->input_channel);
+			input_process->input_channel = NULL;
 
-		mp_server_remove_process (input_process->server, input_process);
-		process_set_status (input_process, MP_PROCESS_DEFUNCT);
+			mp_server_remove_process (input_process->server, input_process);
+			process_set_status (input_process, MP_PROCESS_DEFUNCT);
 		
-		return FALSE;
-	} else {
-		StackElement *stack = NULL;
+			return FALSE;
+		} else {
+			StackElement *stack = NULL;
 
-		if (info.operation == MI_MALLOC ||
-		    info.operation == MI_REALLOC ||
-		    info.operation == MI_FREE) {
-			void **stack_buffer = NULL;
-			StackStash *stash = get_stack_stash (input_process);
-			
-			stack_buffer = g_alloca (sizeof (void *) * info.alloc.stack_size);
-			g_io_channel_read (source, (char *)stack_buffer, sizeof(void *) * info.alloc.stack_size, &count);
-			stack = stack_stash_store (stash, stack_buffer, info.alloc.stack_size);
+//			fprintf (stderr, "Read size %d op 0x%x\n", count, info.operation);
 
-		} else if (info.operation == MI_EXIT) {
-			process_set_status (input_process, MP_PROCESS_EXITING);
-			if (input_process->clone_of)
-				process_detach (input_process);
-		}
+			if (info.operation == MI_MALLOC ||
+			    info.operation == MI_REALLOC ||
+			    info.operation == MI_FREE ||
+			    info.operation == MI_TIME) {
+				void **stack_buffer = NULL;
+				StackStash *stash = get_stack_stash (input_process);
+
+				g_assert (count >= sizeof (MIInfoAlloc));
+
+				stack_buffer = g_alloca (sizeof (void *) * info.alloc.stack_size);
+				g_io_channel_read_chars (source, (char *)stack_buffer, sizeof(void *) * info.alloc.stack_size, &count, NULL);
+				stack = stack_stash_store (stash, stack_buffer, info.alloc.stack_size);
+
+			} else if (info.operation == MI_EXIT) {
+				process_set_status (input_process, MP_PROCESS_EXITING);
+				if (input_process->clone_of)
+					process_detach (input_process);
+			}
 		
-		process = input_process;
-		while (process->clone_of)
-			process = process->clone_of;
+			process = input_process;
+			while (process->clone_of)
+				process = process->clone_of;
 		
-		process_command (process, &info, stack);
+			process_command (process, &info, stack);
 
 
 /*		if (info.any.pid != input_process->pid)
 		g_warning ("Ow! Ow! Ow: %d %d %d!", info.any.pid, input_process->pid, g_io_channel_unix_get_fd (input_process->input_channel)); */
 
-	}
+		}
+	} while (info.operation != MI_EXIT && (g_io_channel_get_buffer_condition(source) & G_IO_IN));
   
 	return TRUE;
 }
@@ -644,10 +673,42 @@ process_start_input (MPProcess *process)
 	if (!process->input_tag && process->input_channel)
 		process->input_tag = g_io_add_watch_full (process->input_channel,
 							  G_PRIORITY_LOW,
-							  G_IO_IN | G_IO_HUP,
+							  G_IO_IN | G_IO_PRI | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
 							  input_func, process, NULL);
 }
 
+void
+process_clear_input (MPProcess *process)
+{
+	g_return_if_fail (process != NULL);
+  
+	process->bytes_used = 0;
+	process->n_allocations = 0;
+
+	if (process->map_list) {
+		g_list_foreach (process->map_list, (GFunc)g_free, NULL);
+		g_list_free (process->map_list);
+      
+		process->map_list = NULL;
+	}
+
+	if (process->bad_pages) {
+		g_list_free (process->bad_pages);
+		process->bad_pages = NULL;
+	}
+
+	if (process->block_table) {
+		g_hash_table_foreach (process->block_table, process_free_block, NULL);
+		g_hash_table_destroy (process->block_table);
+		process->block_table = NULL;
+	}
+
+	if (process->stack_stash) {
+		stack_stash_free (process->stack_stash);
+		process->stack_stash = NULL;
+	}
+}
+
 char *
 process_find_exec (char **args)
 {
@@ -905,7 +966,7 @@ process_detach (MPProcess *process)
 			char response = 0;
 			write (fd, &response, 1);
 		} else {
-			g_io_channel_close (process->input_channel);
+			g_io_channel_shutdown (process->input_channel, TRUE, NULL);
 			process_set_status (process, MP_PROCESS_DETACHED);
 		}
 	}
@@ -951,3 +1012,71 @@ process_block_foreach (MPProcess                 *process,
 	g_hash_table_foreach (block_table, block_table_foreach_func, &info);
 }
 
+static gboolean
+my_str_equal (const char *s1, const char *s2)
+{
+	if (s1 == NULL && s2 == NULL)
+		return TRUE;
+
+	if (s1 == NULL || s2 == NULL)
+		return FALSE;
+
+	return strcmp (s1, s2) == 0;
+}
+
+gboolean
+symbol_equal (gconstpointer s1, gconstpointer s2)
+{
+	const Symbol *sym1 = s1;
+	const Symbol *sym2 = s2;
+
+	if (s1 == NULL && s2 == NULL)
+		return TRUE;
+
+	if (s1 == NULL || s2 == NULL)
+		return FALSE;
+
+	return (my_str_equal (sym1->name, sym2->name)) && (sym1->addr == sym2->addr);
+}
+
+guint
+symbol_hash  (gconstpointer s)
+{
+	const Symbol *symbol = s;
+
+	if (!s)
+		return 0;
+	
+	return symbol->name? g_str_hash (symbol->name) : 0 + symbol->addr;
+}
+
+Symbol *
+symbol_copy  (const Symbol *orig)
+{
+	Symbol *symbol;
+
+	if (!orig)
+		return NULL;
+
+	symbol = g_new0 (Symbol, 1);
+
+	symbol->name = g_strdup (orig->name);
+	symbol->addr = orig->addr;
+	symbol->size = orig->size;
+
+	g_assert (symbol_equal (symbol, orig));
+	g_assert (symbol_hash (symbol) == symbol_hash (orig));
+	
+	return symbol;
+}
+
+void
+symbol_free  (Symbol *symbol)
+{
+	if (!symbol)
+		return;
+
+	g_free (symbol->name);
+	g_free (symbol);
+}
+
